/*
 * Copyright (C) 2001-2002 WarKnite Computers.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 */

#include "struct.h"

CHANUSER register_chanuser(char *name, char *pass, NICKUSER founder, time_t created)
{
    Chanuser *n, *c;
    n = (CHANUSER) malloc(sizeof(Chanuser));
    if (n == NULL) return NULL;

#ifdef HASHWATCH
    myhash.chanuser++;
#endif

    memset(n, 0, sizeof(Chanuser));
    c = cshead;
    if (n != NULL)
    {

       n->name = (char *)malloc(strlen(name+1));
       if (n->name) strcpy(n->name,name);
       else
       {
           debug("Error","Error allocating chanuser");
           free(n);
           return;
       }
       strcpy(n->pass,pass);
       n->registered = created;
       n->founder = founder;
       n->successor = NULL;

       if (cshead == NULL)
       {
           cshead = n;
           n->next = NULL;
           n->prev = NULL;
       }
       else
       {
           cshead->prev = n;
           n->prev = NULL;
           n->next = cshead;
           cshead = n;
       }
    }
    return n;
}

void delete_chanuser(CHANUSER c)
{
    int i=0;

    if (!c->prev && c->next)
    {
       c->next->prev = NULL;
       cshead = c->next;
    }
    else if (c->next && c->prev)
    {
       c->next->prev = c->prev;
       c->prev->next = c->next;
    }
    else if (c->prev && !c->next) c->prev->next = NULL;
    else cshead = NULL;
    if (c->name) free(c->name);
#ifdef HASHWATCH
    myhash.nickuser--;
#endif
    c = NULL;
}

CHANUSER find_chanuser(char *name)
{
   Chanuser *c;
   c = cshead;

   assert(name);
   while (c)
   {
      if (!strcasecmp(name,c->name)) return c;
      c = c->next;
   }
   return NULL;
}

int reg_chan(char *name, char *pass, char *desc)
{
   Link *ptr;

   if (!name) notice(curuser->nick,"Usage: REGISTER <channel> <pass> :<desc>");
   else if (name[0] != '#') notice (curuser->nick,"%s is not a valid channel name.",name);
   else if (!(curnick = find_nickuser(curuser->nick)))
   {
      notice(curuser->nick,"You must register your nickname first. "
              "\2/msg %s HELP\2 for information on registering nicknames.",service->nick);
   }
   else if (!HasIdentified(curnick)) notice(curuser->nick,"Please identify with \2%s\2 first.",nickserv->nick);
   else if (now - curuser->last_chanreg < CHAN_DELAY)
   { 
      curuser->last_chanreg += 5;

      notice(curuser->nick,
          "Please wait %d seconds before using the REGISTER command again.",
              CHAN_DELAY-(now-curuser->last_chanreg));
   }
   else if ((cschan = find_chanuser(name)))
   {
/*      if (IsForbidden(cschan)) notice(curuser->nick,"Channel \2%s\2 may not be registered.",name);
 *      else
 */
       notice (curuser->nick,"Channel \2%s\2 already registered!",name);
   }
   else if (!is_chanop(curuser,name)) notice(curuser->nick,"You must be a channel operator to register the channel.");
   else if (curnick->channelcount >= MAX_CHANS) notice(curuser->nick,"Sorry, you have already have the limit of %d channels.",MAX_CHANS);
   else if (!strcasecmp(curuser->nick,pass) || !strcasecmp(name,pass) || !strcasecmp(name+1,pass))
   {
      notice(curuser->nick,"Your password is too easy.",MIN_PASSLEN);
   }
   else
   {
      cschan = register_chanuser(name,pass,curnick,now);
      cschan->lastidentify = curnick;
      makepass(cschan->pass,pass);
      cschan->desc = strdup(desc);
      cschan->url = strdup("");
      cschan->lastused = now;
      curnick->channelcount++;

      debug("Register","Channel %s registered by %s!%s@%s",name,curuser->nick,curuser->user,curuser->host);

      notice(curuser->nick,
              "Channel \2%s\2 registered under your nickname: \2%s\2",
              name,curuser->nick);

      notice(curuser->nick,
              "Your channel password is \2%s\2 - write it down and "
              "keep it in a safe place.  For your own protection, do "
              "not give this password to anyone.",pass);

   }
}

int parse_chanserv(PARAMS)
{
   if (!strcasecmp(argv[0],"help")) notice(curuser->nick,"Chanserv help not yet implemented.");
   else if (!strcasecmp(argv[0],"register"))
   {
      if (argc >= 4) reg_chan(argv[1],argv[2],argv[3]);
      else reg_chan(NULL,NULL,NULL);
   }
   else notice(curuser->nick,"Unknown command %s. Type \"/msg %s HELP\" for help.",argv[0],service->nick);

}
