#ifndef DEFINE_H_
#define DEFINE_H_

#define MAXFDS                  1024
#define PARAMS                  int argc, char *argv[]
#define RANDOM(min,max)         (min + ((max - min) * (rand() / (RAND_MAX + 1.0))))

/* Service types */
#define S_SECURITY              1
#define S_CHANSERV              2              
#define S_NICKSERV              3
#define S_OPERSERV              4
#define S_MEMOSERV              5
#define S_HELDNICK              6
#define S_JUPED			7

/* Socket types */
#define SOCK_OFFLINE            -3        /* Disconnected from uplink */
#define SOCK_DELETE             -2        /* Delete from the connection list */
#define SOCK_INVAILD            -1        /* Invaild Socket */
#define SOCK_UPLINK             1         /* Uplink server */
#define SOCK_RESUME             99999

/* Connection status */
#define CN_FAILED	        -1        /* Failed to connect */
#define CN_TRYING               1         /* Trying to Connect */
#define CN_CONNECTED            2         /* connected sent pass/server */
#define CN_LOGGED               3         /* got pass reply */

/* Sizes */
#define NICKLEN                 9
#define USERLEN                 10
#define REALLEN                 50
#define HOSTLEN                 63
#define PASSLEN                 20
#define PASSBUF                 PASSLEN+1

#define TOPICLEN                160
#define CHANNELLEN              200
#define KEYLEN                  23
#define USERHOST                NICKLEN+USERLEN+HOSTLEN+3

/* User Flags */
#define FLAGS_OPER              0x1
#define FLAGS_ADMIN             0x2
#define FLAGS_INVISIBLE         0x4
#define FLAGS_WALLOPS           0x8
#define FLAGS_AWAY              0x10

#define IsAnOper(x)             ((x)->flags & FLAGS_OPER)
#define IsAdmin(x)              ((x)->flags & FLAGS_ADMIN)
#define IsInvisible(x)          ((x)->flags & FLAGS_INVISIBLE)
#define IsWallops(x)            ((x)->flags & FLAGS_WALLOPS)
#define IsAway(x)               ((x)->flags & FLAGS_AWAY)

#define SetOper(x)              ((x)->flags |= FLAGS_OPER)
#define SetAdmin(x)             ((x)->flags |= FLAGS_ADMIN)
#define SetInvisible(x)         ((x)->flags |= FLAGS_INVISIBLE)
#define SetWallops(x)           ((x)->flags |= FLAGS_WALLOPS)
#define SetAway(x)              ((x)->flags |= FLAGS_AWAY)

#define ClearOper(x)            ((x)->flags &= ~FLAGS_OPER)
#define ClearAdmin(x)           ((x)->flags &= ~FLAGS_ADMIN)
#define ClearInvisible(x)       ((x)->flags &= ~FLAGS_INVISIBLE)
#define ClearWallops(x)         ((x)->flags &= ~FLAGS_WALLOPS)
#define ClearAway(x)            ((x)->flags &= ~FLAGS_AWAY)

/* Nickname flags */
#define NICK_ADMIN              0x1
#define NICK_IDENTIFIED         0x2
#define NICK_RECOGNIZED         0x4
#define NICK_FORBIDDEN          0x8
#define NICK_OPER		0x10
#define NICK_SECURE		0x20
#define NICK_KILLPROT		0x40

#define IsServAdmin(x)          ((x)->modes & NICK_ADMIN)
#define HasIdentified(x)        ((x)->modes & NICK_IDENTIFIED)
#define IsRecognized(x)         ((x)->modes & NICK_RECOGNIZED)
#define IsForbidden(x)          ((x)->modes & NICK_FORBIDDEN)
#define IsOperProt(x)		((x)->modes & NICK_OPER)
#define IsSecure(x)             ((x)->modes & NICK_SECURE)
#define IsKillProt(x)           ((x)->modes & NICK_KILLPROT)

#define SetServAdmin(x)         ((x)->modes |= NICK_ADMIN)
#define SetIdentified(x)        ((x)->modes |= NICK_IDENTIFIED)
#define SetRecognized(x)        ((x)->modes |= NICK_RECOGNIZED)
#define SetForbidden(x)         ((x)->modes |= NICK_FORBIDDEN)
#define SetOperProt(x)          ((x)->modes |= NICK_OPER)
#define SetSecure(x)            ((x)->modes |= NICK_SECURE)
#define SetKillProt(x)          ((x)->modes |= NICK_KILLPROT)

#define ClearServAdmin(x)       ((x)->modes &= ~NICK_ADMIN)
#define ClearIdentified(x)      ((x)->modes &= ~NICK_IDENTIFIED)
#define ClearRecognized(x)      ((x)->modes &= ~NICK_RECOGNIZED)
#define ClearForbidden(x)       ((x)->modes &= ~NICK_FORBIDDEN)
#define ClearOperProt(x)        ((x)->modes &= ~NICK_OPER)
#define ClearSecure(x)          ((x)->modes &= ~NICK_SECURE)
#define ClearKillProt(x)        ((x)->modes &= ~NICK_KILLPROT)

/* Channel Flags */

#define MODE_PRIVATE            0x1
#define MODE_SECRET             0x2
#define MODE_MODERATED          0x4
#define MODE_TOPICLIMIT         0x8
#define MODE_INVITEONLY         0x10
#define MODE_NOPRIVMSGS         0x20
#define MODE_ANONOPS		0x40
#define MODE_KEY                0x80
#define MODE_BAN       	        0x100
#define MODE_LIMIT              0x200
#define MODE_OPERATOR           0x400
#define MODE_VOICE              0x800
#define MODE_HALFOP             0x1000

#define IsPrivate(x)            ((x)->modes & MODE_PRIVATE)
#define IsSecret(x)             ((x)->modes & MODE_SECRET)
#define IsModerated(x)          ((x)->modes & MODE_MODERATED)
#define IsTopicLimit(x)         ((x)->modes & MODE_TOPICLIMIT)
#define IsInviteOnly(x)         ((x)->modes & MODE_INVITEONLY)
#define IsNoPrivmsgs(x)         ((x)->modes & MODE_NOPRIVMSGS)
#define IsKeyed(x)              ((x)->modes & MODE_KEY)
#define IsLimited(x)            ((x)->modes & MODE_LIMIT)
#define IsOped(x)               ((x)->modes & MODE_OPERATOR)
#define IsVoice(x)              ((x)->modes & MODE_VOICE)
#define IsHalfop(x)             ((x)->modes & MODE_HALFOP)

#define SetPrivate(x)           ((x)->modes |= MODE_PRIVATE)
#define SetSecret(x)            ((x)->modes |= MODE_SECRET)
#define SetModerated(x)         ((x)->modes |= MODE_MODERATED)
#define SetTopicLimit(x)        ((x)->modes |= MODE_TOPICLIMIT)
#define SetInviteOnly(x)        ((x)->modes |= MODE_INVITEONLY)
#define SetNoPrivmsgs(x)        ((x)->modes |= MODE_NOPRIVMSGS)
#define SetKeyed(x)             ((x)->modes |= MODE_KEY)
#define SetLimited(x)           ((x)->modes |= MODE_LIMIT)
#define SetOped(x)              ((x)->modes |= MODE_OPERATOR)
#define SetVoice(x)             ((x)->modes |= MODE_VOICE)
#define SetHalfop(x)            ((x)->modes |= MODE_HALFOP)

#define ClearPrivate(x)         ((x)->modes &= ~MODE_PRIVATE)
#define ClearSecret(x)          ((x)->modes &= ~MODE_SECRET)
#define ClearModerated(x)       ((x)->modes &= ~MODE_MODERATED)
#define ClearTopicLimit(x)      ((x)->modes &= ~MODE_TOPICLIMIT)
#define ClearInviteOnly(x)      ((x)->modes &= ~MODE_INVITEONLY)
#define ClearNoPrivmsgs(x)      ((x)->modes &= ~MODE_NOPRIVMSGS)
#define ClearKeyed(x)           ((x)->modes &= ~MODE_KEY)
#define ClearLimited(x)         ((x)->modes &= ~MODE_LIMIT)
#define ClearOped(x)            ((x)->modes &= ~MODE_OPERATOR)
#define ClearVoice(x)           ((x)->modes &= ~MODE_VOICE)
#define ClearHalfop(x)          ((x)->modes &= ~MODE_HALFOP)

#endif
