/*
 * Copyright (C) 2001-2002 WarKnite Computers.
 * Timeout stuff taken from Services NN-1.3 1988 release version.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 */

#include "struct.h"
#include "timeout.h"

#define TO_COLLIDE      0       /* Collide the user with this nick */
#define TO_RELEASE      1       /* Release a collided nick */

static void timeout_collide(Timeout *t);
static void timeout_release(Timeout *t);

static struct my_timeout
{
  struct my_timeout *next, *prev;
  Nickuser *ni;
  Timeout *to;
  int type;
} *my_timeouts;

NICKUSER register_nickuser(char *nick, char *pass, time_t created, time_t lastonline)
{
    Nickuser *n, *c;
    n = (NICKUSER) malloc(sizeof(Nickuser));
    if (n == NULL) return NULL;

#ifdef HASHWATCH
    myhash.nickuser++;
#endif

    memset(n, 0, sizeof(Nickuser));
    c = nickhead;
    if (n != NULL)
    {

       n->nick = (char *)malloc(strlen(nick+1));
       if (n->nick) strcpy(n->nick,nick);
       else
       {
           debug("Error","Error allocating nickuser");
           free(n);
           return;
       }
       strcpy(n->pass,pass);
       n->created = created;
       n->lastonline = lastonline;

       if (nickhead == NULL)
       {
           nickhead = n;
           n->next = NULL;
           n->prev = NULL;
       }
       else
       {
           nickhead->prev = n;
           n->prev = NULL;
           n->next = nickhead;
           nickhead = n;
       }
    }
    return n;
}

void delete_nickuser(NICKUSER c)
{
    int i=0;

    if (!c->prev && c->next)
    {
       c->next->prev = NULL;
       nickhead = c->next;
    }
    else if (c->next && c->prev)
    {
       c->next->prev = c->prev;
       c->prev->next = c->next;
    }
    else if (c->prev && !c->next) c->prev->next = NULL;
    else nickhead = NULL;
    if (c->nick) free(c->nick);
    if (c->lastinfo) free(c->lastinfo);
    if (c->lastmask) free(c->lastmask);
    if (c->access)
    {
      for (i = 0; i < c->accesscount; i++) free(c->access[i]);
      free(c->access);
    }
    free(c);
#ifdef HASHWATCH
    myhash.nickuser--;
#endif
    c = NULL;
}

static void add_ns_timeout(Nickuser *ni, int type, time_t delay)
{
   Timeout *to;
   struct my_timeout *t;
   void (*timeout_routine) (Timeout *);

   if (type == TO_COLLIDE) timeout_routine = timeout_collide;
   else if (type == TO_RELEASE) timeout_routine = timeout_release;
   else return;

   to = add_timeout(delay, timeout_routine);
   to->data = ni;
   t = malloc(sizeof (*t));
   t->next = my_timeouts;
   my_timeouts = t;
   t->prev = NULL;
   t->ni = ni;
   t->to = to;
   t->type = type;
}

static void del_ns_timeout(Nickuser *ni, int type)
{
   struct my_timeout *t, *t2;

   t = my_timeouts;
   while (t)
   {
      if (t->ni == ni && t->type == type)
      {
         t2 = t->next;
         if (t->next) t->next->prev = t->prev;
         if (t->prev) t->prev->next = t->next;
         else my_timeouts = t->next;

         del_timeout (t->to);
         free (t);
         t = t2;
      }
      else t = t->next;
   }
}

static void rem_ns_timeout(Nickuser *ni, int type)
{
   struct my_timeout *t, *t2;

   t = my_timeouts;
   while (t)
   {
      if (t->ni == ni && t->type == type)
      {
         t2 = t->next;
         if (t->next) t->next->prev = t->prev;

         if (t->prev) t->prev->next = t->next;
         else my_timeouts = t->next;
         free (t);
         t = t2;
      }
      else t = t->next;
   }
}

NICKUSER find_nickuser(char *nick)
{
   Nickuser *c;
   c = nickhead;

   assert(nick);
   while (c)
   {
      if (!strcasecmp(nick,c->nick)) return c;
      c = c->next;
   }
   return NULL;
}

static void timeout_enforce(Timeout *t)
{
   SERVICES heldnick;
   USERPTR user;
   char *nick = t->data;
   
   heldnick = register_service(nick,"enforcer",me->name,"Nick Enforcer","+i",S_HELDNICK);
   user = find_nick(nick);

   if (user)
   {
      service_login(heldnick,curconn,user->timestamp-1);
      purge_user(user);
   }
   else service_login(heldnick,curconn,now);

   join_channel(nick,"#dump");

   free (nick);
}

static void collide(Nickuser *ni, int from_timeout)
{
   Timeout *to;
   void (*timeout_routine) (Timeout *);

   if (!from_timeout) del_ns_timeout(ni, TO_COLLIDE);

   add_ns_timeout(ni, TO_RELEASE, RELEASE_TIMEOUT);

   timeout_routine = timeout_enforce;
   to = add_timeout(15, timeout_routine);
   to->data = strdup(ni->nick);
}

static void timeout_collide(Timeout *t)
{
  Nickuser *ni = t->data;
  User *u;

  rem_ns_timeout(ni, TO_COLLIDE);
  if (HasIdentified(ni) || !(u = find_nick(ni->nick)) || u->signon > t->settime) return;
  collide(ni, 1);
}

static void release(Nickuser *ni, int from_timeout)
{
   SERVICES heldnick;
   if (!from_timeout) del_ns_timeout(ni, TO_RELEASE);

   heldnick = find_service(ni->nick);
   if (!heldnick) debug("Release","Couldn't find heldnick %s, sending quit anyways",ni->nick);
   else delete_service(heldnick);

   send_to_client(curconn->fd,":%s QUIT :Released\n",ni->nick);
}

static void timeout_release(Timeout *t)
{
   Nickuser *ni = t->data;
   rem_ns_timeout(ni, TO_RELEASE);
   release(ni, 1);
}

void update_user(USERPTR user)
{
   if (user && (curnick = find_nickuser(user->nick)))
   {
      ClearIdentified(curnick);
      ClearRecognized(curnick);
      del_ns_timeout(curnick, TO_COLLIDE);
   }
}

int is_on_access(USERPTR user, NICKUSER nick)
{
   int i;
   char *buf;

   buf = malloc(strlen(user->user) + strlen(user->host) + 2);
   sprintf(buf, "%s@%s", user->user, user->host);

   for (i = 0; i < nick->accesscount; i++)
   {
      if (match_wild(nick->access[i], buf))
      {
         free(buf);
         return 1;
      }
   }
   free(buf);
   return 0;
}

int validate_user(USERPTR user)
{
   int access=0;

   if (!(curnick = find_nickuser(user->nick))) return 0;

   if (IsForbidden(curnick))
   {
       notice (user->nick,
           "This nickname may not be used. Please choose another nick to use. "
           "If you do not change nicks within one minute, you will be forced "
           "off the network.");
       add_ns_timeout(curnick,TO_COLLIDE,60);     
       return 0;
   }
   
   access = is_on_access(user,curnick);

   if (IsAnOper(user) && IsOperProt(curnick) && access)
   {
      SetIdentified(curnick);
      if (!IsRecognized(curnick))
      {
         curnick->lastonline = now;
         if (curnick->lastmask) free(curnick->lastmask);
         curnick->lastmask = malloc(strlen(user->user) + strlen(user->host) + 2);
         sprintf(curnick->lastmask,"%s@%s",user->user,user->host);
         if (curnick->lastinfo) free(curnick->lastinfo);
         curnick->lastinfo = strdup(user->info);
      }
      notice (user->nick,"You are now recognized.");
      return 1;
   }

   if (IsSecure(curnick))
   {
      notice (user->nick,
              "This nickname is registered and protected.  If it is your "
              "nick, type \2/msg %s IDENTIFY \37password\37\2.  Otherwise, "
              "please choose a different nick.", service->nick);
      if (access && !IsOperProt(curnick)) return 1;
   }
   else
   {
      if (access && !IsOperProt(curnick))
      {
         SetRecognized(curnick);
         curnick->lastonline = now;
         if (curnick->lastmask) free(curnick->lastmask);
         curnick->lastmask = malloc(strlen(user->user) + strlen(user->host) + 2);
         sprintf(curnick->lastmask,"%s@%s",user->user,user->host);
         if (curnick->lastinfo) free(curnick->lastinfo);
         curnick->lastinfo = strdup(user->info);
         return 1;
      }
      else
      {
          notice (user->nick,
             "This nick is owned by someone else.  Please choose another.  "
             "(If this is your nick, type \2/msg %s IDENTIFY \37password\37\2.)",service->nick);
      }
   }

   if (!IsOperProt(curnick) && !IsSecure(curnick))
   {
      notice (user->nick,
              "To add your current hostmask to your access list, "
              "type \2/msg %s ACCESS ADD\2 now.",service->nick);
   }
   
   if (IsKillProt(curnick) || IsOperProt(curnick))
   {
      notice(user->nick,
         "If you do not change nicks within one minute, you will be "
         "forced off the network.");

      add_ns_timeout(curnick,TO_COLLIDE,60);
   }
   return 0;
}

static int reg_nick(char *nick, char *pass)
{
   char temp[512];
   if (curnick = find_nickuser(nick))
   {
      if (IsForbidden(curnick)) notice(nick,"This nick may not be used.");
      else notice(nick,"Your nick is already registered.");
   }
   else if (!pass) notice(nick,"Usage: REGISTER <password>");
   else if (strlen(pass) > MAX_PASSLEN) notice(nick,"Password is too long.");
   else if (match_wild("guest*",nick)) notice(nick,"You can't register guest nicks.");
   else if ((now - curuser->last_nickreg) < NICK_DELAY) notice(nick,"Please wait %lu seconds before registering another nick.",NICK_DELAY-(now-curuser->last_nickreg));
   else if ((!strcasecmp(pass,curuser->nick) || (!strcasecmp(pass,curuser->user)) || (strlen(pass) < MIN_PASSLEN)))
   {
      notice(nick,"Your password is too easy.",MIN_PASSLEN);
   }
   else
   {
       curnick = register_nickuser(nick,encrypt(pass),now,curuser->signon);
       debug("Nick","%s registered by %s@%s.",nick,curuser->user,curuser->host,pass);

       curnick->lastonline = curuser->signon;
       curnick->lastmask = malloc(strlen(curuser->user) + strlen(curuser->host) + 2);
       sprintf (curnick->lastmask,"%s@%s",curuser->user,curuser->host);
       curnick->lastinfo = strdup(curuser->info);

       curnick->accesscount = 1;
       curnick->access = malloc(sizeof(char *));
       curnick->access[0] = create_mask(curuser);
       SetIdentified(curnick);

       notice(nick,"Nickname %s registered under host: %s.",curnick->nick,curnick->access[0]);
       curuser->last_nickreg = now;
   }
}

static int drop_nick(char *nick, char *why)
{
   int operdrop=0;
   if (strcasecmp(nick,curuser->nick))
   {
      if (!IsAnOper(curuser))
      {
         notice(curuser->nick,"You can't drop other people's nicknames.");
         return;
      }
      else operdrop = 1;
   }
   curnick = find_nickuser(nick);
   if (!curnick)
   {
      if (!operdrop) notice(curuser->nick,"Your nick isn't registered.");
      else notice(curuser->nick,"%s isn't registered.",nick);
   }
   else if (!HasIdentified(curnick) && !operdrop) notice(curuser->nick,"You need to identify before using this command.");
   else if (IsServAdmin(curnick)) notice(curuser->nick,"You can't drop an admin nick.");
   else
   {
      if (operdrop && !why)
      {
         notice(curuser->nick,"Usage: DROP <nick> :<reason>");
         return;
      }
      else
      {
         if (operdrop) debug("Drop","Oper %s!%s@%s dropped nickname %s.",curuser->nick,curuser->user,curuser->host,nick);
         else debug("Drop","%s!%s@%s dropped nickname %s.",nick,curuser->user,curuser->host,nick);
         notice(curuser->nick,"Nickname %s has been dropped.",nick);
         delete_nickuser(curnick);
      }
   }
}

static int identify_nick(char *nick, char *pass)
{
   curnick = find_nickuser(nick);
   if (!curnick) notice(nick,"Your nick isn't registered.");
   else if (HasIdentified(curnick)) notice(nick,"You are already identified.");
   else if (!pass) notice(nick,"Usage: IDENTIFY <password>");
   else
   {
      if (checkpass(pass,curnick->pass))
      {
         debug("Identify","%s!%s@%s identified for nickname %s.",curuser->nick,curuser->user,curuser->host,nick);
         SetIdentified(curnick);
         notice(nick,"Password accepted - you are now recognized.");

      }
      else
      {
         debug("Identify","%s!%s@%s invalid password for nickname %s.",curuser->nick,curuser->user,curuser->host,nick);
         notice(nick,"Invalid password for nickname %s.",nick);
         bad_pass(curuser);
      }
   }
}

static int ghost_nick(char *nick, char *pass)
{
   char temp[512];
   if (!nick) notice(curuser->nick,"Usage: GHOST <nick> <pass>");
   else if (!(curnick = find_nickuser(nick))) notice(curuser->nick,"Nick %s isn't registered.",nick);
   else if (!find_nick(nick)) notice(curuser->nick,"Nick %s isn't currently in use.",nick);
   else if (!strcasecmp(nick,curuser->nick)) notice(curuser->nick,"You can't ghost yourself!");
   else if (find_service(nick)) notice(curuser->nick,"Nick %s can't be ghosted at this time.\n",nick);
   else
   {
      if (!checkpass(pass,curnick->pass))
      {
         debug("Ghost","%s!%s@%s invalid password for nickname %s.",curuser->nick,curuser->user,curuser->host,nick);
         notice(curuser->nick,"Invalid password for nickname %s.",nick);
         bad_pass(curuser);
      }
      else
      {
         sprintf(temp,"GHOST command used by %s.",curuser->nick);
         do_kill(service->nick,nick,temp);
         notice(curuser->nick,"The ghost with your nick has been killed.");
      }
   }
}

static int forbid_nick(char *nick)
{
   if (!IsAnOper(curuser)) 
   {
      notice(curuser->nick,"Unknown command.");
      return;
   }
   else if (!nick)
   {
      notice(curuser->nick,"Usage: FORBID <nick>");
      return;
   }

   if (curnick = find_nickuser(nick))
   {
      if (IsServAdmin(curnick))
      { 
         notice(curuser->nick,"You can't forbid an admin nickname.");
         return;
      }
      else delete_nickuser(curnick);
   }
   
   if (curnick = register_nickuser(nick,"",now,now))
   {
      SetForbidden(curnick);
      debug("Forbid","%s!%s@%s set FORBID on nick %s",curuser->nick,curuser->user,curuser->host,nick);
      notice(curuser->nick,"Nick %s is now Forbidden.",nick);
   }
   else
   {
      debug("Forbid","Error setting FORBID on nick %s",nick);
   }
}

static int info_nick(char *nick)
{
   struct tm *thetime;
   char buffer[512];

   if (!nick) notice(curuser->nick,"Usage: INFO <nick>");
   else if (!(curnick = find_nickuser(nick))) notice(curuser->nick,"Nick %s isn't registered.",nick);
   else if (IsForbidden(curnick)) notice(curuser->nick,"Nick %s is forbidden.",nick);
   else
   {
      thetime = localtime(&curnick->created);
      strftime(buffer, sizeof(buffer), "%b%e %T %G", thetime);
      notice(curuser->nick,"Info for %s:",nick);
      notice(curuser->nick,"              Info: %s",curnick->lastinfo);
      notice(curuser->nick,"      Last address: %s",curnick->lastmask);
      notice(curuser->nick,"   Time Registered: %s",buffer);
      thetime = localtime(&curnick->lastonline);
      strftime(buffer, sizeof(buffer), "%b%e %T %G", thetime);
      notice(curuser->nick,"         Last seen: %s",buffer);

      memset(buffer,0,sizeof(buffer));
      if (IsOperProt(curnick)) strcat(buffer, "Oper protection");
      if (IsKillProt(curnick))
      {
         if (*buffer) strcat(buffer, ", ");
         strcat(buffer, "Kill Protection");
      }
      if (IsSecure(curnick))
      {
         if (*buffer) strcat(buffer, ", ");
         strcat(buffer, "Security");
      }
      if (!*buffer) strcpy(buffer, "None");
      notice(curuser->nick,"           Options: %s",buffer);
   }
}

static int recover_nick(char *nick, char *pass)
{
   char temp[512];
   if (!nick) notice(curuser->nick,"Usage: RECOVER <nick> <pass>");
   else if (!(curnick = find_nickuser(nick))) notice(curuser->nick,"Nick %s isn't registered.",nick);
   else if (!find_nick(nick)) notice(curuser->nick,"Nick %s isn't currently in use.",nick);
   else if (!strcasecmp(nick,curuser->nick)) notice(curuser->nick,"You can't recover yourself!");
   else
   {
      if (!checkpass(pass,curnick->pass))
      {
         debug("Recover","%s!%s@%s invalid password for nickname %s.",curuser->nick,curuser->user,curuser->host,nick);
         notice(curuser->nick,"Invalid password for nickname %s.",nick);
         bad_pass(curuser);
      }
      else
      {
         collide(curnick, 0);
         notice(curuser->nick,"User claiming your nick has been killed.");
         notice(curuser->nick,
              "\2/msg %s RELEASE\2 to get it back before the one-minute "
              "timeout.", service->nick);
      }
   }
}

static int release_nick(char *nick, char *pass)
{
   SERVICES heldnick;
   char temp[512];
   if (!nick) notice(curuser->nick,"Usage: RELEASE <nick> <pass>");
   else if (!(curnick = find_nickuser(nick))) notice(curuser->nick,"Nick %s isn't registered.",nick);
   else
   {
      heldnick = find_service(nick);
      if (!heldnick) notice(curuser->nick,"Nick %s isn't currently in use.",nick);
      else if (heldnick->type != S_HELDNICK) notice(curuser->nick,"Nick %s isn't currently in use.",nick);
      else
      {
         if (!checkpass(pass,curnick->pass))
         {
            debug("Release","%s!%s@%s invalid password for nickname %s.",curuser->nick,curuser->user,curuser->host,nick);
            notice(curuser->nick,"Invalid password for nickname %s.",nick);
            bad_pass(curuser);
         }
         else
         {
            release(curnick,0);
            notice(curuser->nick,"Services' hold on your nick has been released.");
         }
      }
   }
}

static int set_nick_pass(char *nick, char *pass, char *why)
{
   NICKUSER temp;
   if (strlen(pass) > MAX_PASSLEN) notice(nick,"Password is too long.");
   else if ((!strcasecmp(pass,curuser->nick) || (!strcasecmp(pass,curuser->user)) || (strlen(pass) < MIN_PASSLEN)))
   {
      notice(nick,"Your password is too easy.",MIN_PASSLEN);
   }
   else if (why)  /* Opers setting a pass */
   {
      if (!(curnick = find_nickuser(nick))) notice(curuser->nick,"%s is not registered.",nick);
      else if (IsForbidden(curnick)) notice(curuser->nick,"%s is forbidden.",nick);
      else if (IsServAdmin(curnick))
      {
         temp = find_nickuser(curuser->nick);
         if (!IsServAdmin(temp)) notice(curuser->nick,"Admin passwords can only be changed by the admin.");
         else
         {
            strcpy(curnick->pass,encrypt(pass));
            notice(curuser->nick,"Password for %s changed.",curnick->nick);
         }
      }
      else
      {
         strcpy(curnick->pass,encrypt(pass));
         notice(curuser->nick,"Password for %s changed.",curnick->nick);
      }
   }
   else
   {
      strcpy(curnick->pass,encrypt(pass));
      notice(curuser->nick,"Password changed.");
   }
}

static int do_set_kill(NICKUSER nick, int option)
{
   switch(option)
   {
      case -1:
         notice(curuser->nick,"Usage: SET KILL <ON|OFF>");
         break;
      
      case 0:
         SetKillProt(nick);
         notice(curuser->nick,"Kill protection is now ON");
         break;
      
      case 1:
         ClearKillProt(nick);
         notice(curuser->nick,"Kill protection is now OFF");
         break;
   }
}

static int do_set_oper(NICKUSER nick, int option)
{
   if (!IsAnOper(curuser))
   {
      notice(curuser->nick,"Unknown SET option OPER");
      return;
   }

   switch(option)
   {
      case -1:
         notice(curuser->nick,"Usage: SET OPER <ON|OFF>");
         break;

      case 0:
         SetOperProt(nick);
         notice(curuser->nick,"Oper protection is now ON");
         break;

      case 1:
         ClearOperProt(nick);
         notice(curuser->nick,"Oper protection is now OFF");
         break;
   }
}

static int do_set_secure(NICKUSER nick, int option)
{
   switch(option)
   {
      case -1:
         notice(curuser->nick,"Usage: SET SECURE <ON|OFF>");
         break;

      case 0:
         SetSecure(nick);
         notice(curuser->nick,"Secure option is now ON");
         break;

      case 1:
         ClearSecure(nick);
         notice(curuser->nick,"Secure option is now OFF");
         break;
   }
}

static int parse_nickset(PARAMS)
{
   int tmp=-1;

   curnick = find_nickuser(curuser->nick);
   if (!curnick) notice(curuser->nick,"Your nickname isn't registered.");
   else if (!HasIdentified(curnick) && !IsAnOper(curuser)) notice(curuser->nick,"Password authentication required for that command.");
   else
   {
      if (argc >= 2)
      {
         if (!strcasecmp(argv[1],"ON")) tmp = 0;
         else if (!strcasecmp(argv[1],"OFF")) tmp = 1;
      }

      if (!strncasecmp(argv[0],"pass",4))
      {
         if (IsAnOper(curuser))
         {
            if (argc == 2) set_nick_pass(curuser->nick,argv[1],NULL);
            else if (argc >= 4) set_nick_pass(argv[2],argv[1],argv[3]);
            else notice(curuser->nick,"Usage: SET PASSWORD <password> <nick> :<why>");
         }
         else
         {
            if (argc >= 2) set_nick_pass(curuser->nick,argv[1],NULL);
            else notice(curuser->nick,"Usage: SET PASSWORD <password>");
         }
         return;
      }
      else if (!strcasecmp(argv[0],"kill")) do_set_kill(curnick,tmp);
      else if (!strcasecmp(argv[0],"oper")) do_set_oper(curnick,tmp);
      else if (!strcasecmp(argv[0],"secure")) do_set_secure(curnick,tmp);
      else notice(curuser->nick,"Unknown SET option %s.",argv[0]);
   }
}

static int parse_nickaccess(PARAMS)
{
   char **access;
   char *mask;
   int i,mademask=0;
   
   if (!strcasecmp(argv[0],"list") && (argc == 2) && IsAnOper(curuser))
   {
      if (!(curnick = find_nickuser(argv[1])))
      {
         notice(curuser->nick,"Nick %s isn't registered.",argv[1]);
         return;
      }
      else
      {
         notice(curuser->nick,"Access list for %s:",argv[1]);
         for (access = curnick->access, i=0; i < curnick->accesscount; access++, i++)
         {
             notice(curuser->nick,"%d. %s",i+1,*access);
         }
      }  
   }
   else if (argc < 2 && !strcasecmp(argv[0],"del")) notice(curuser->nick,"Usage: ACCESS <ADD|DEL|LIST> [hostmask]");
   else if (argc >= 2 && !strchr(argv[1],'@')) notice(curuser->nick,"Mask must be in the form user@host.");
   else if (!(curnick = find_nickuser(curuser->nick))) notice(curuser->nick,"Your nick isn't registered.");
   else if (!HasIdentified(curnick)) notice(curuser->nick,"Password authentication required for that command.");
   else if (!strcasecmp(argv[0],"list"))
   {
      notice(curuser->nick,"Access list:");
      for (access = curnick->access, i=0; i < curnick->accesscount; access++, i++)
      {
         notice(curuser->nick,"%d. %s",i+1,*access);
      }
   }
   else if (!strcasecmp(argv[0],"add"))
   {
      if (curnick->accesscount >= MAX_ACCESS)
      {
         notice(curuser->nick,"Your access list reached the max number of entries.");
         return;
      }

      if (argc < 2)
      {
         mask = create_mask(curuser);
         mademask=1;
      }
      else mask = argv[1];

      for (access = curnick->access, i=0; i < curnick->accesscount; access++, i++)
      {
         if (!strcasecmp(*access,mask))
         {
            notice(curuser->nick,"Mask %s already present in your access list.",*access);
            if (mademask) free(mask);
            return;
         }
      }
      curnick->accesscount++;
      curnick->access = realloc(curnick->access, sizeof(char *) * curnick->accesscount);
      curnick->access[curnick->accesscount - 1] = strdup(mask);
      notice(curuser->nick,"%s added to your access list.",mask);
      if (mademask) free(mask);
      return;
   }
   else if (!strcasecmp(argv[0],"del"))
   {
      mask = argv[1];
      for (access = curnick->access,i=0; i < curnick->accesscount; access++,i++)
      {
         if (!strcasecmp(*access, mask))
         {
            notice(curuser->nick,"%s deleted from your access list.",*access);
            free (*access);
            curnick->accesscount--;
            if (i < curnick->accesscount) bcopy (access + 1, access, (curnick->accesscount - i) * sizeof (char *));
            if (curnick->accesscount) curnick->access = realloc (curnick->access,curnick->accesscount * sizeof (char*));
            else
            {
               free(curnick->access);
               curnick->access = NULL;
            }
            return;
         }
      }
      notice(curuser->nick,"%s not found in your access list.",mask);
   }
   else notice(curuser->nick,"Usage: ACCESS <ADD|DEL|LIST> [hostmask]");

}

static int list_nick(char *pattern)
{
   int nicks=0,i=0;
   char buf[512];
   if (!pattern) notice(curuser->nick,"Usage: INFO <nick|pattern>");
   else
   {
      notice(curuser->nick,"List of entries matching %s:",pattern);
      for (curnick = nickhead; curnick; curnick = curnick->next)
      {
         if (IsForbidden(curnick)) continue;
         snprintf (buf, sizeof (buf), "%-20s  %s",curnick->nick,curnick->lastmask);

         if (!strcasecmp(pattern,curnick->nick) || match_wild(pattern,curnick->nick))
         {
            if (nicks++ <= 15) notice(curnick->nick,"     %d. %s",nicks,buf);
            else nicks++;
         }
      }
      notice(curuser->nick,"End of list - %d/%d matches shown.",nicks > 15 ? 15:nicks,nicks);
   }
}

int parse_nickserv(PARAMS)
{
   if (!strcasecmp(argv[0],"help")) notice(curuser->nick,"Nickserv help not yet implemented.");
   else if (!strcasecmp(argv[0],"register"))
   {
      if (argc >= 2) reg_nick(curuser->nick,argv[1]);
      else reg_nick(curuser->nick,NULL);
   }
   else if (!strcasecmp(argv[0],"drop"))
   {
      if (argc >= 3) drop_nick(argv[1],argv[2]);
      else if (argc == 2) drop_nick(argv[1],"");
      else drop_nick(curuser->nick,NULL);
   }
   else if (!strcasecmp(argv[0],"identify"))
   {
      if (argc >= 2) identify_nick(curuser->nick,argv[1]);
      else identify_nick(curuser->nick,NULL);
   }
   else if (!strcasecmp(argv[0],"ghost"))
   {
      if (argc < 3) ghost_nick(NULL,NULL);
      else ghost_nick(argv[1],argv[2]);
   }
   else if (!strcasecmp(argv[0],"recover"))
   {
      if (argc < 3) recover_nick(NULL,NULL);
      else recover_nick(argv[1],argv[2]);
   }
   else if (!strcasecmp(argv[0],"release"))
   {
      if (argc < 3) release_nick(NULL,NULL);
      else release_nick(argv[1],argv[2]);
   }
   else if (!strcasecmp(argv[0],"info"))
   {
      if (argc >= 2) info_nick(argv[1]);
      else info_nick(NULL);
   }
   else if (!strcasecmp(argv[0],"set"))
   {
      if (argc >= 2) parse_nickset(argc-1,argv+1);
      else notice(curuser->nick,"Usage: SET <option> <parameters>"); 
   }
   else if (!strcasecmp(argv[0],"access"))
   {
      if (argc >= 2) parse_nickaccess(argc-1,argv+1);
      else notice(curuser->nick,"Usage: ACCESS <ADD|DEL|LIST> [hostmask]");
   }
   else if (IsAnOper(curuser))
   {
      if (!strcasecmp(argv[0],"forbid"))
      {
         if (argc >= 2) forbid_nick(argv[1]);
         else forbid_nick(NULL);
      }
      else if (!strcasecmp(argv[0],"list"))
      {
         if (argc >= 2) list_nick(argv[1]);
         else list_nick(NULL);
      }
      else notice(curuser->nick,"Unknown command %s. Type \"/msg %s HELP\" for help.",argv[0],service->nick);
   }
  else notice(curuser->nick,"Unknown command %s. Type \"/msg %s HELP\" for help.",argv[0],service->nick);
}
