/*
 * Copyright (C) 2001-2002 WarKnite Computers.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 */

#include "struct.h"

int parse_umode(char *source, char *umode);

int parse_server(PARAMS)
{
    SERVPTR uplink;
    char *up = argv[0];
    char *server = argv[2];
    char *info = argv[4];

    curserv = find_server(server);
    if (!curserv)
    {
       uplink = find_server(up);
       if (!uplink)
       {
          debug("Error","Server Uplinked to nonexistant server %s",argv[0]);
          return;
       }
       curserv = register_server(server,info,atoi(argv[3]),uplink);
       assert(curserv);
       register_downlink(curserv,uplink);
    }
    debug("Server","%s linked to %s (%s)",server,up,info);
}

int parse_nick(PARAMS)
{
    time_t ts;
    char *nick = argv[1];
    char *ident = argv[5];
    char *host = argv[6];
    char *server = argv[7];
    char *info = argv[8];

    curuser = find_nick(nick);
    if (curuser)
    {
       debug("Error","User %s already exists, deleting older",nick);
       delete_user(curuser);
    }

    sscanf(argv[3],"%lu",&ts);
    curuser = register_user(nick,ident,host,info);
    curuser->timestamp = ts;
    curserv = find_server(server);
    parse_umode(nick,argv[4]);

    if (curserv)
    {
       curuser->server = curserv;
       register_servuser(curserv,curuser);
    }
    else debug("Error","User %s on nonexistant server %s",curuser->nick,server);
    debug("Nick","%s!%s@%s (%s) connected to %s",nick,ident,host,info,server);

    service = nickserv;
    validate_user(curuser);       
}

int parse_nickchange(PARAMS)
{
    time_t ts;
    char *oldnick=argv[0];
    char *newnick=argv[2];
    
    curuser = find_nick(oldnick);
    if (curuser)
    {
       sscanf(argv[3],"%lu",&ts);
       strcpy(curuser->nick,newnick);
       curuser->timestamp = ts;
       debug("Nick","%s is now known as %s",oldnick,newnick);
    }
    else
    { 
       debug("Error","Nick change for nonexistant user %s",oldnick);            
    }

    curnick = find_nickuser(oldnick);
    if (curnick) update_user(curuser);

    service = nickserv;  
    validate_user(curuser);
}

int parse_chanusers(char *buffer)
{
   USERPTR user;
   Link *lp;
   char temp[512], *argv[255];
   int argc=0,x=0;
  
   strcpy(temp,buffer);
   tokenize(temp, argv, 255, &argc, 1);

   for (x = 0; x < argc; x++)
   {
      switch(argv[x][0])
      {
         case '@':
            if (argv[x][1] == '@')
            {
               argv[x]+=2;
            }
            else argv[x]++;

            user = find_nick(argv[x]);
            if (user)
            {
                lp = register_chan_user(curchan,user);
                SetOped(lp);
                lp = register_user_chan(user,curchan);
                SetOped(lp);
            }
            debug("Join","%s Joined %s (with ops)",argv[x],curchan->name);
            break;

         case '+':
            argv[x]++;
            user = find_nick(argv[x]);
            if (user)
            {
                lp = register_chan_user(curchan,user);
                SetVoice(lp);
                lp = register_user_chan(user,curchan);
                SetVoice(lp);
            }
            debug("Join","%s Joined %s (with voice)",argv[x],curchan->name);
            break;

         case '%':
            argv[x]++;
            user = find_nick(argv[x]);
            if (user)
            {
                lp = register_chan_user(curchan,user);
                SetHalfop(lp);
                lp = register_user_chan(user,curchan);
                SetHalfop(lp);
            }
            debug("Join","%s Joined %s (with halfop)",argv[x],curchan->name);
            break;

         case '\0':     /* Kludge (some stuff does : @nick) (space gets converted to \0) */
            break;

         default:
            user = find_nick(argv[x]);
            if (user)
            {
                register_chan_user(curchan,user);
                register_user_chan(user,curchan);
            }
            debug("Join","%s Joined %s",argv[x],curchan->name);
            break;
      }
   }
}

int parse_sjoin(PARAMS)
{
    time_t ts;
    int x=0,y=0,mode=0;

    curchan = find_channel(argv[3]);

    if (!curchan)
    {
       sscanf(argv[2],"%lu",&ts);
       curchan = register_channel(argv[3]);
       curchan->timestamp = ts;
    }

    y=5;
    for (x = 0; argv[4][x]; x++)
    {
       switch (argv[4][x])
       {
          case '+':
             mode = 1;
             break;
 
          case '-':
             mode = 0;
             break;

          case 'k':
             y++;
             break;

          case 'l':
             y++;
             break;

       }
    }
    parse_chanusers(argv[y]);
}

int parse_part(PARAMS)
{
   Link *lp;
   char *nick=argv[0];
   char *chan=argv[2];

   curchan = find_channel(chan);
   curuser = find_nick(nick);

   if (!curchan)
   {
      debug("Error","part from nonexistant channel %s",chan);
      return;
   }
   else if (!curuser)
   {
      debug("Error","part from nonexistant user %s",nick);
      return;
   }
   else
   {
      delete_chan_user(curchan,curuser);
      delete_user_chan(curuser,curchan);
   }

   if (!curchan->userhead)
   {
      debug("Part","%s is now empty",chan);
      delete_channel(curchan);
   }

   debug("Part","%s has left %s",nick,chan);
}

int parse_kick(PARAMS)
{
   Link *lp;
   char *nick=argv[3];
   char *chan=argv[2];

   curchan = find_channel(chan);
   curuser = find_nick(nick);
 
   if (!curchan)
   {
      debug("Error","Kick from nonexistant channel %s",chan);
      return;
   }
   else if (!curuser)
   {
      debug("Error","Kick from nonexistant user %s",nick);
      return;
   }
   else
   {
      delete_chan_user(curchan,curuser);
      delete_user_chan(curuser,curchan);
   }

   if (!curchan->userhead)  /* user may kick themself */
   {
      debug("Part","%s is now empty",chan);
      delete_channel(curchan);
   }

   debug("Kick","%s kicked %s from %s (%s)",argv[0],nick,chan,argv[4]);
}

int parse_privmsg(PARAMS)
{
   char *nick=argv[0];
   char temp[512];
   static char *aargv[255];
   int aargc=0;
   
   curuser = find_nick(nick);

   strcpy(temp,argv[3]);
   tokenize(temp, aargv, 255, &aargc, 0);
 
   switch(argv[2][0])
   {
      case '#':
      case '&':
          return;
          break;
      
      default:
          break;
   }

   if (!(service = find_service(argv[2]))) return;

   if (argv[3][0] == '\1')
   {
      argv[3]++;
      if (argv[3][strlen(argv[3])-1] == '\1') argv[3][strlen(argv[3])-1] = '\0';
      if (!strncasecmp(argv[3],"PING",4)) notice(nick,"\1%s\1",argv[3]);
      return;
   }

   switch(service->type)
   {
      case S_SECURITY:
         parse_security(aargc,aargv);
         break;

      case S_OPERSERV:
         parse_operserv(aargc,aargv);
         break;
            
      case S_NICKSERV:
         parse_nickserv(aargc,aargv);
         break;
           
      case S_CHANSERV:
         parse_chanserv(aargc,aargv);
         break;

      case S_MEMOSERV:
         parse_memoserv(aargc,aargv);
         break;
      
      default:
         break;

   }
}

int parse_quit(PARAMS)
{
   char *nick=argv[0];

   curuser = find_nick(nick);
   if (curuser) purge_user(curuser);
   else debug("Error","Quit for nonexistant user %s",nick);

   debug("Quit","%s has quit irc",nick);
   curnick = find_nickuser(nick);
   if (curnick)
   {
      update_user(curuser);
      curnick->lastonline = now;
   }
}

int parse_kill(PARAMS)
{
   char *nick=argv[2];

   curuser = find_nick(nick);
   if (curuser)
   {
      purge_user(curuser);
      debug("Kill","Recieved kill for %s",nick);
   }
   else
   {
      service = find_service(nick);
      if (service)
      {
         service_login(service,curconn,now);
         if (service == security) join_channel(service->nick,"#thefactory");
         join_channel(service->nick,"#dump");
      }
      else debug("Error","Kill for nonexistant user %s",nick);
   }
   curnick = find_nickuser(nick);

   if (curnick)
   {
      ClearRecognized(curnick);
      ClearIdentified(curnick);
      curnick->lastonline = now;
   }
}

int parse_umode(char *source, char *umode)
{
   int x=0,mode=0;

   for (x = 0; umode[x]; x++)
   {
      switch (umode[x])
      {
          case '+':
             mode = 1;
             break;

          case '-':
             mode = 0;
             break;

          case 'o':
             debug("Usermode","%s is %s an operator. (IRCop)",source,mode?"now":"no longer");
             if (mode) SetOper(curuser);
             else ClearOper(curuser);
             break;

          case 'a':
             debug("Usermode","%s is %s an operator. (Admin)",source,mode?"now":"no longer");
             if (mode) SetAdmin(curuser);
             else ClearAdmin(curuser);
             break;

          case 'w':
             if (mode) SetWallops(curuser);
             else ClearWallops(curuser);
             break;

          case 'i':
             if (mode) SetInvisible(curuser);
             else ClearInvisible(curuser);
             break;

          default:
             debug("Mode","Unknown mode %s%c",mode?"+":"-",umode[x]);
             break;
      }
   }
}

int parse_cmode(PARAMS)
{
   Link *user, *chan;
   char *source=argv[0];

   int x=0,y=4,mode=0;

   curchan = find_channel(argv[2]);
   if (!curchan)
   {
      debug("Error","Mode for nonexistant channel (%s)",argv[2]);
      return;
   }

   y=4;
   for (x = 0; argv[3][x]; x++)
   {
      switch (argv[3][x])
      {
          case '+':
             mode = 1;
             break;

          case '-':
             mode = 0;
             break;

          case 'o':
             curuser = find_nick(argv[y]);
             if (!curuser)
             {
                debug("Error","Op for nonexistant user %s",argv[y]);
                continue;;
             }

             user = find_chan_link(curuser,curchan);
             chan = find_user_link(curchan,curuser);
             if (!user) debug("Error","Channel %s not in %s's list",curchan->name,curuser->nick);
             else if (!chan) debug("Error","User %s not in %s's list",curuser->nick,curchan->name);
             else
             {
                if (mode)
                {
                   SetOped(user);
                   SetOped(chan);
                }
                else
                {
                   ClearOped(user);
                   ClearOped(chan);
                }
             }
             y++;
             break;

          case 'v':
             curuser = find_nick(argv[y]);
             if (!curuser)
             {
                debug("Error","voice for nonexistant user %s",argv[y]);
                continue;
             }
             user = find_chan_link(curuser,curchan);
             chan = find_user_link(curchan,curuser);
             if (!user) debug("Error","Channel %s not in %s's list",curchan->name,curuser->nick);
             else if (!chan) debug("Error","User %s not in %s's list",curuser->nick,curchan->name);
             else
             {
                if (mode)
                {
                   SetVoice(user);
                   SetVoice(chan);
                }
                else
                {
                   ClearVoice(user);
                   ClearVoice(chan);
                }
             }
             y++;
             break;

          case 'h':
             curuser = find_nick(argv[y]);
             if (!curuser)
             {
                debug("Error","Halfop for nonexistant user %s",argv[y]);
                continue;
             }
             user = find_chan_link(curuser,curchan);
             chan = find_user_link(curchan,curuser);

             if (!user) debug("Error","Channel %s not in %s's list",curchan->name,curuser->nick);
             else if (!chan) debug("Error","User %s not in %s's list",curuser->nick,curchan->name);
             else
             {
                if (mode)
                {
                   SetHalfop(user);
                   SetHalfop(chan);
                }
                else
                {
                   ClearHalfop(user);
                   ClearHalfop(chan);
                }
             }
             y++;
             break;

          case 'k':
          case 'b':
          case 'e':
          case 'I':
             y++;
             break;

          case 'l':
             if (mode) y++;
             break;

          case 'n':
             if (mode) SetNoPrivmsgs(curchan);
             else ClearNoPrivmsgs(curchan);
             break;

          case 't':
             if (mode) SetTopicLimit(curchan);
             else ClearTopicLimit(curchan);
             break;

          case 's':
             if (mode) SetSecret(curchan);
             else ClearSecret(curchan);
             break;

          case 'p':
             if (mode) SetPrivate(curchan);
             else ClearPrivate(curchan);
             break;

          case 'm':
             if (mode) SetModerated(curchan);
             else ClearModerated(curchan);
             break;

          case 'i':
             if (mode) SetInviteOnly(curchan);
             else ClearInviteOnly(curchan);
             break;

          default:
             debug("Mode","Unknown mode %s%c",mode?"+":"-",argv[3][x]);
             break;
      }
   }
}
                    
int parse_mode(PARAMS)
{
   char *source=argv[0];
   char buffer[512];
   int temp=0;

   curuser = find_nick(source);
   if (!curuser && !index(source,'.'))
   {
      debug("Error","Mode from nonexistant user (%s)",source);
      return;
   }

   memset(buffer,0,sizeof(buffer));
   for (temp = 3; temp < argc; temp++)
   {
      strcat(buffer,argv[temp]);
      strcat(buffer," ");
   }
   if (argv[2][0] == '#') debug("Mode","%s MODE %s %s",source,argv[2],buffer);

   if (!strcasecmp(source,argv[2])) parse_umode(source,argv[3]);
   else parse_cmode(argc,argv);

}

int parse_split(SERVPTR serv)
{
   SERVPTR down;
   Link *lp;

   while (serv->down)
   {
      if (serv->down->next) lp = serv->down->next;
      else lp = NULL;

      down = serv->down->value.serv;
      if (down == me)
      {
         serv->down = lp;
         continue;
      }
      
      debug("Split","Server %s has split from %s",down->name,serv->name);

      cleanup_users(down);

      parse_split(down);
      delete_server(down);

      free(serv->down);
      serv->down = lp;
   }
   serv->down = NULL;
}

int parse_squit(PARAMS)
{
   char *server=argv[1];
   char *uplink=argv[2];
   
   curserv = find_server(server);
   if (curserv)
   {
      parse_split(curserv);

      cleanup_users(curserv);
      debug("Split","Server %s has split from %s",curserv->name,curserv->up->name);
      delete_downlink(curserv,curserv->up);
      delete_server(curserv);
   }
}

int parse_error(PARAMS)
{
   debug("Error","I disconnected from my uplink");
   if (me->up)
   {
      parse_split(me->up);
      cleanup_users(me->up);
      debug("Split","Server %s has split from %s",me->name,me->up->name);
      delete_downlink(me,me->up);
      delete_server(me->up);
      me->up = NULL;
   }
}

int parse_eob(PARAMS)
{
   join_channel("Security","#thefactory");
   services_join("#dump");
   burst=0;
}

int parse_wallops(PARAMS)
{
  char *who=argv[0];
  char *msg=argv[2];
  debug("Wallops","!%s! %s",who,msg);
}

int parse_operwall(PARAMS)
{
  char *who=argv[0];
  char *msg=argv[2];
  debug("Operwall","!%s! %s",who,msg);
}

int parse_away(PARAMS)
{
  char *nick=argv[0];

  if (argc == 3)
  {
     curuser = find_nick(nick);
     if (curuser)
     {
        SetAway(curuser);
        debug("Away","User %s is now away (%s)",nick,argv[2]);
     }
     else
     {
        debug("Error","Away from nonexistant user %s",nick);
     }
  }
  else
  {
     curuser = find_nick(nick);
     if (curuser)
     {
        ClearAway(curuser);
        debug("Away","User %s is now back from away",nick);
     }
     else
     {
        debug("Error","Away from nonexistant user %s",nick);
     }
  }
}
