/*
 * Copyright (C) 2001-2002 WarKnite Computers.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 */

#include "struct.h"

void privmsg(char *chan, char *format, ...)
{
   va_list va; char buffer[1024];
   va_start(va,format); vsnprintf(buffer,1024,format,va); va_end(va);
   send_to_client(curconn->fd,":%s PRIVMSG %s :%s\n",security->nick,chan,buffer);
}

void notice(char *target, char *format, ...)
{
   va_list va; char buffer[1024];
   va_start(va,format); vsnprintf(buffer,1024,format,va); va_end(va);

   if (!service) service = security;
   send_to_client(curconn->fd,":%s NOTICE %s :%s\n",service->nick,target,buffer);
}

SERVICES find_service(char *nick)
{
   Services *c;
   c = serviceshead;

   assert(nick);
   while (c)
   {
      if (!strcasecmp(nick,c->nick)) return c;
      c = c->next;
   }
   return NULL;
}

SERVICES register_service(char *nick, char *user, char *host, char *info, char *modes, u_short type)
{
    Services *n, *c;
    n = (SERVICES) malloc(sizeof(Services));
    if (n == NULL) return NULL;

#ifdef HASHWATCH
    myhash.services++;
#endif

    memset(n, 0, sizeof(Services));
    c = serviceshead;
    if (n != NULL)
    {

       n->nick = (char *)malloc(strlen(nick)+1);
       n->user = (char *)malloc(strlen(user)+1);
       n->host = (char *)malloc(strlen(host)+1);
       n->info = (char *)malloc(strlen(info)+1);
       n->modes = (char *)malloc(strlen(modes)+1);
       n->type = type;

       if (!n->nick || !n->user || !n->host || !n->info || !n->modes)
       {
          debug("Error","allocating memory for service");
          if (n->nick) free(n->nick);
          if (n->user) free(n->user);
          if (n->host) free(n->host);
          if (n->info) free(n->info);
          if (n->modes) free(n->modes);

          free(n);
          return;
       }
       strcpy(n->nick,nick);
       strcpy(n->user,user);
       strcpy(n->host,host);
       strcpy(n->info,info);
       strcpy(n->modes,modes);

       if (serviceshead == NULL)
       {
           serviceshead = n;
           n->next = NULL;
           n->prev = NULL;

       }
       else
       {
           serviceshead->prev = n;
           n->prev = NULL;
           n->next = serviceshead;
           serviceshead = n;
       }
    }
    return n;
}

void delete_service(SERVICES c)
{
    if (!c->prev && c->next)
    {
       c->next->prev = NULL;
       serviceshead = c->next;
    }
    else if (c->next && c->prev)
    {
       c->next->prev = c->prev;
       c->prev->next = c->next;
    }
    else if (c->prev && !c->next) c->prev->next = NULL;
    else serviceshead = NULL;

    if (c->nick) free(c->nick);
    if (c->user) free(c->user);
    if (c->host) free(c->host);
    if (c->info) free(c->info);
    if (c->modes) free(c->modes);
    free(c);
#ifdef HASHWATCH
    myhash.services--;
#endif
    c = NULL;
}

int bad_pass(USERPTR user)
{
    user->attempts++;
    if (user->attempts > MAX_TRYS-1)
    {
       do_kill(service->nick,user->nick,"Too many invalid passwords.");
    }
}

int join_channel(char *nick, char *channel)
{
   curchan = find_channel(channel);
   if (curchan) send_to_client(curconn->fd,":%s SJOIN %lu %s + :@%s\n",me->name,curchan->timestamp,channel,nick);
   else send_to_client(curconn->fd,":%s SJOIN %lu %s + :@%s",me->name,now,channel,nick);
}

int services_join(char *channel)
{
   for(service = serviceshead; service; service = service->next)
   {
      join_channel(service->nick,channel);
   }
}

int service_login(SERVICES service, CONNPTR conn, time_t ts)
{
   send_to_client(conn->fd,"NICK %s 0 %lu %s %s %s %s :%s\n",service->nick,ts,service->modes,service->user,service->host,me->name,service->info);
}

int make_services(CONNPTR conn)
{
   for (service = serviceshead; service; service = service->next) service_login(service,conn,now);
}

int init_services()
{
   security = register_service("Security","security",me->name,"Security services","+Soki",S_SECURITY);
   register_service("Operserv2","operserv2",me->name,"Operator services","+Soki",S_OPERSERV);
   nickserv = register_service("Nickserv2","nickserv2",me->name,"Nickname services","+Soki",S_NICKSERV);
   register_service("Chanserv2","Chanserv2",me->name,"Channel services","+Sokid",S_CHANSERV);
   register_service("Memoserv2","memoserv2",me->name,"Memo services","+Soki",S_MEMOSERV);
}
