/*
 * Copyright (C) 2001-2002 WarKnite Computers.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 */

#include <arpa/inet.h>
#include <netdb.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/param.h>
#include <ctype.h>
#include <sys/ioctl.h>
#include <net/if.h>
#include "struct.h"

void send_to_client(int thefd, char *format, ...)
{
   CONNPTR conn;
   int bytes=0;
   va_list va; char buffer[1024];
   va_start(va,format); vsnprintf(buffer,1024,format,va); va_end(va);
   if (buffer != NULL) bytes = write(thefd,buffer,strlen(buffer));
   if (bytes < 0)
   {
       conn = find_fd(thefd);
       if (!conn) return;
       if (errno == EPIPE) kill_user(conn);
   }
}

static int sock_read(CONNPTR conn, int count)
{
  int bytes;

  while (conn->buffer[conn->bufindex] != '\n')
  {
      bytes = recv(conn->fd, conn->buffer + conn->bufindex, 1, 0);

      if (bytes <= 0)
      {
          if ((errno == EWOULDBLOCK) || (errno == EAGAIN)) return SOCK_RESUME;
          return -1;
      }
      else if (conn->buffer[conn->bufindex] == '\r');
      else if ((conn->bufindex<count) && (conn->buffer[conn->bufindex]!='\n')) conn->bufindex++;
      else
      {
          conn->buffer[conn->bufindex] = '\0';
          break;
      }
  }
  return conn->bufindex;
}

int SockOpts(void)
{
    int     s,pram;

    if ((s = socket(AF_INET,SOCK_STREAM,0)) < 0) return(-1);

    pram = 0;
    setsockopt(s,SOL_SOCKET,SO_LINGER,(char*)&pram,sizeof(int));
    pram = 1;
    setsockopt(s,SOL_SOCKET,SO_KEEPALIVE,(char*)&pram,sizeof(int));
    fcntl(s,F_SETFL,O_NONBLOCK | fcntl(s,F_GETFL));

    return(s);
}

void setnonblocking(int sock)
{
    int opts;

    opts = fcntl(sock,F_GETFL);
    if (opts < 0)
    {
       perror("fcntl(F_GETFL)");
       exit(0);
    }

    opts = (opts | O_NONBLOCK);
    if (fcntl(sock,F_SETFL,opts) < 0)
    {
       perror("fcntl(F_SETFL)");
       exit(0);
    }
    return;
}

unsigned long get_ip(char *host)
{
    struct  hostent *hp;
    struct  in_addr *in;
    unsigned long   ip;

    if ((ip = inet_addr(host)) == -1)
    {
        hp = gethostbyname(host);
        if (!hp) return(-1);
        in = (struct in_addr*)(hp->h_addr_list[0]);
        ip = (unsigned long)(in->s_addr);
        return(ip);
    }
    return(ip);
}

int SockConnect(char *host, int port)
{
        struct  sockaddr_in sai;
        int     s,uv;

        if ((s = SockOpts()) < 0) return(SOCK_INVAILD);

        if (!host) return -1;
        uv = 0;

        memset((char*)&sai,0,sizeof(struct sockaddr_in));
        sai.sin_family = AF_INET;
        sai.sin_port = htons(port);
        if ((sai.sin_addr.s_addr = get_ip(host)) == -1) return(-1);

        uv = connect(s,(struct sockaddr*)&sai,sizeof(sai));

        if ((uv < 0) && (errno != EINPROGRESS))
        {
            close(s);
            return(-1);
        }
        return(s);
}

int kill_user(CONNPTR conn)
{
   debug("Connect","Connection lost: FD=%d",conn->fd);
   switch (conn->type)
   {
       case SOCK_UPLINK:
           (void)close(conn->fd);
           conn->fd = SOCK_OFFLINE;
           conn->status = 0;
           burst=1;
           break;
           
       default:
           break;
   }
}

int deal_with_data(CONNPTR conn)
{
   int bytes;

   if ((bytes = sock_read(conn,511)) < 0)
   {
      switch(errno)
      {
         case EINTR:
         case EAGAIN:
            break;

         default:
            kill_user(conn);
      }
   }
   else if ((bytes != SOCK_RESUME) && (bytes != 0))
   {
      switch (conn->type)
      {
         case SOCK_UPLINK:
             process_uplink(conn);
             break;

         default: 
             break;
      }
      conn->bufindex = 0;
      memset(conn->buffer, 0, sizeof(conn->buffer));
   }
}

void connect_to_uplink(CONNPTR conn)
{
   conn->lasttry = now;

   if ((conn->fd = SockConnect("bub.necrohell.net",5556)) > 0) conn->status = CN_TRYING;
   else conn->status = CN_FAILED;
}
