/*
 * Copyright (C) 2001-2002 WarKnite Computers.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 */

#include "struct.h"

CONNPTR register_connection(int fd, int type)
{
    Connection *n, *c;
    n = (CONNPTR) malloc(sizeof(Connection));
    if (n == NULL) return NULL;

#ifdef HASHWATCH
    myhash.connections++;
#endif

    memset(n, 0, sizeof(Connection));
    c = connhead;
    if (n != NULL)
    {

       n->fd = fd;
       n->type = type;

       if (connhead == NULL)
       {
           connhead = n;
           n->next = NULL;
           n->prev = NULL;

       }
       else
       {
           connhead->prev = n;
           n->prev = NULL;
           n->next = connhead;
           connhead = n;
       }
    }
    return n;
}

void delete_connection(CONNPTR c)
{
    if (!c->prev && c->next)
    {
       c->next->prev = NULL;
       connhead = c->next;
    }
    else if (c->next && c->prev)
    {
       c->next->prev = c->prev;
       c->prev->next = c->next;
    }
    else if (c->prev && !c->next) c->prev->next = NULL;
    else connhead = NULL;
    free(c);
#ifdef HASHWATCH
    myhash.connections--;
#endif
    c = NULL;
}

CHANPTR register_channel(char *name)
{
    Channel *n, *c;
    n = (CHANPTR) malloc(sizeof(Channel));
    assert(n);

#ifdef HASHWATCH
    myhash.channels++;
#endif
    memset(n, 0, sizeof(Channel));
    c = chanhead;
    if (n != NULL)
    {

       n->name = (char *)malloc(strlen(name)+1);
       if (n->name) strcpy(n->name,name);
       else
       {
           debug("Error","Error allocating channel memory");
           free(n);
           return NULL;
       }
       n->userhead = NULL;

       if (chanhead == NULL)
       {
           chanhead = n;
           n->next = NULL;
           n->prev = NULL;
       }
       else
       {
           chanhead->prev = n;
           n->next = chanhead;
           n->prev = NULL;
           chanhead = n;
       }

    }
    return n;
}

void delete_channel(CHANPTR c)
{
    if (!c->prev && c->next)
    {
       c->next->prev = NULL;
       chanhead = c->next;
    }
    else if (c->next && c->prev)
    {
       c->next->prev = c->prev;
       c->prev->next = c->next;
    }
    else if (c->prev && !c->next) c->prev->next = NULL;
    else chanhead = NULL;
    if (c->name) free(c->name);
    free(c);
    c = NULL;
#ifdef HASHWATCH
    myhash.channels--;
#endif
}

USERPTR register_user(char *nick, char *user, char *host, char *info)
{
    User *n, *c;
    n = (USERPTR) malloc(sizeof(User));
    assert(n);

#ifdef HASHWATCH
    myhash.users++;
#endif

    memset(n, 0, sizeof(User));
    c = userhead;
    if (n != NULL)
    {

       strcpy(n->host,host);
       strcpy(n->nick,nick);
       strcpy(n->user,user);
       n->info = (char *)malloc(strlen(info)+1);
       if (n->info) strcpy(n->info,info);
       else
       {
          debug("Error","Unable to malloc info (%d bytes)\n",strlen(info)+1);
          free(n);
          return NULL;
       }
       n->last_nickreg = now-60;
       n->last_chanreg = now-60;
       n->signon = now;

       n->server = NULL;
       n->chanhead = NULL;         

       if (userhead == NULL)
       {
           userhead = n;
           n->next = NULL;
           n->prev = NULL;

       }
       else
       {
           userhead->prev = n;
           n->next = userhead;
           n->prev = NULL;
           userhead = n;
       }
    }
    return n;
}

void delete_user(USERPTR c)
{
    if (!c->prev && c->next)
    {
       c->next->prev = NULL;
       userhead = c->next;
    }
    else if (c->next && c->prev)
    {
       c->next->prev = c->prev;
       c->prev->next = c->next;
    }
    else if (c->prev && !c->next) c->prev->next = NULL;
    else userhead = NULL;
    if (c->info) free(c->info);
    free(c);
    c = NULL;
#ifdef HASHWATCH
    myhash.users--;
#endif
}

SERVPTR register_server(char *name, char *info, int hops, SERVPTR uplink)
{
    Server *n, *c;
    n = (SERVPTR) malloc(sizeof(Server));
    if (n == NULL) return;
#ifdef HASHWATCH
    myhash.servers++;
#endif
    memset(n, 0, sizeof(Server));
    c = servhead;
    if (n != NULL) {

       n->hops = hops;
       strcpy(n->name,name);

       n->info = (char *)malloc(strlen(info)+1);
       if (n->info) strcpy(n->info,info);
       else
       {
          debug("Error","Unable to malloc info (%d bytes)\n",strlen(info)+1);
          free(n);
          return NULL;
       }
       n->up = uplink;
       n->down = NULL;
       n->userhead = NULL;

       if (servhead == NULL)
       {
           servhead = n;
           n->next = NULL;
           n->prev = NULL;
       }
       else
       {
           servhead->prev = n;
           n->next = servhead;
           n->prev = NULL;
           servhead = n;
       }
    }
    return n;
}

void delete_server(SERVPTR c)
{
    if (!c->prev && c->next)
    {
       c->next->prev = NULL;
       servhead = c->next;
    }
    else if (c->next && c->prev)
    {
       c->next->prev = c->prev;
       c->prev->next = c->next;
    }
    else if (c->prev && !c->next) c->prev->next = NULL;
    else servhead = NULL;
    if (c->info) free(c->info);
    free(c);
    c = NULL;
#ifdef HASHWATCH
    myhash.servers--;
#endif

}

Link *make_link()
{
    Link *n;
    n = (Link *)malloc(sizeof(Link));
    assert(n);

#ifdef HASHWATCH
    myhash.links++;
#endif
    memset(n, 0, sizeof(Link));
    return n;
}

Link *register_downlink(SERVPTR down, SERVPTR up)
{
    Link *n, *c;
    n = make_link();
    c = up->down;

    if (n != NULL)
    {
       n->value.serv = down;

       if (up->down == NULL)
       {
           up->down = n;
           n->next = NULL;
           n->prev = NULL;

       }
       else
       {
           up->down->prev = n;
           n->next = up->down;
           n->prev = NULL;
           up->down = n;
       }
    }
    return n;
}

void delete_downlink(SERVPTR down, SERVPTR up)
{
    Link *pNode, *pPrev;
    pNode = up->down;
    pPrev = NULL;

    while(pNode)
    {
         if (pNode->value.serv == down)
         {
              if (pPrev != NULL)
              {
                  pPrev->next = pNode->next;
                  if (pNode->next) pNode->next->prev = pPrev;
              }
              else
              {
                  up->down = pNode->next;
                  if (up->down) up->down->prev = NULL;
              }
              free(pNode);
              pNode = NULL;
#ifdef HASHWATCH
              myhash.links--;
#endif
              break;
         }
         pPrev = pNode;
         pNode = pNode->next;
    }
}

Link *register_servuser(SERVPTR serv, USERPTR user)
{
    Link *n, *c;
    n = make_link();
    c = serv->userhead;

    if (n != NULL)
    {
       n->value.user = user;

       if (serv->userhead == NULL)
       {
           serv->userhead = n;
           n->next = NULL;
           n->prev = NULL;

       }
       else
       {
           serv->userhead->prev = n;
           n->next = serv->userhead;
           n->prev = NULL;
           serv->userhead = n;
       }
    }
    return n;
}

void delete_servuser(SERVPTR serv, USERPTR user)
{
    Link *pNode, *pPrev;
    pNode = serv->userhead;
    pPrev = NULL;

    while(pNode)
    {
         if (pNode->value.user == user)
         {
              if (pPrev != NULL)
              {
                  pPrev->next = pNode->next;
                  if (pNode->next) pNode->next->prev = pPrev;
              }
              else
              {
                  serv->userhead = pNode->next;
                  if (serv->userhead) serv->userhead->prev = NULL;
              }
              free(pNode);
              pNode = NULL;
#ifdef HASHWATCH
              myhash.links--;
#endif
              break;
         }
         pPrev = pNode;
         pNode = pNode->next;
    }
}

Link *register_chan_user(CHANPTR chan, USERPTR user)
{
    Link *n, *c;
    n = make_link();
    c = chan->userhead;

    if (n != NULL)
    {
       n->value.user = user;

       if (chan->userhead == NULL)
       {
           chan->userhead = n;
           n->next = NULL;
           n->prev = NULL;

       }
       else
       {
           chan->userhead->prev = n;
           n->next = chan->userhead;
           n->prev = NULL;
           chan->userhead = n;
       }
    }
    return n;
}

void delete_chan_user(CHANPTR chan, USERPTR user)
{
    Link *pNode, *pPrev;
    pNode = chan->userhead;
    pPrev = NULL;

    while(pNode)
    {
         if (pNode->value.user == user)
         {
              if (pPrev != NULL)
              {
                  pPrev->next = pNode->next;
                  if (pNode->next) pNode->next->prev = pPrev;
              }
              else
              {
                  chan->userhead = pNode->next;
                  if (chan->userhead) chan->userhead->prev = NULL;
              }
              free(pNode);
              pNode = NULL;
#ifdef HASHWATCH
              myhash.links--;
#endif
              break;
         }
         pPrev = pNode;
         pNode = pNode->next;
    }
}

Link *register_user_chan(USERPTR user, CHANPTR chan)
{
    Link *n, *c;
    n = make_link();
    c = user->chanhead;

    if (n != NULL)
    {
       n->value.chan = chan;

       if (user->chanhead == NULL)
       {
           user->chanhead = n;
           n->next = NULL;
           n->prev = NULL;
       }
       else
       {
           user->chanhead->prev = n;
           n->next = user->chanhead;
           n->prev = NULL;
           user->chanhead = n;
       }
    }
    return n;
}

void delete_user_chan(USERPTR user, CHANPTR chan)
{
    Link *pNode, *pPrev;
    pNode = user->chanhead;
    pPrev = NULL;

    while(pNode)
    {
         if (pNode->value.chan == chan)
         {
              if (pPrev != NULL)
              {
                  pPrev->next = pNode->next;
                  if (pNode->next) pNode->next->prev = pPrev;
              }
              else
              {
                  user->chanhead = pNode->next;
                  if (user->chanhead) user->chanhead->prev = NULL;
              }
              free(pNode);
              pNode = NULL;
#ifdef HASHWATCH
              myhash.links--;
#endif
              break;
         }
         pPrev = pNode;
         pNode = pNode->next;
    }
}

CONNPTR find_fd(int fd)
{
   Connection *c;
   c = connhead;
   while (c)
   {
      if (c->fd == fd) return c;
      c = c->next;
   }
   return NULL;
}

CHANPTR find_channel(char *name)
{
   Channel *c;
   c = chanhead;

   assert(name);
   while (c)
   {
      if (!strcasecmp(name,c->name)) return c;
      c = c->next;
   }
   return NULL;
}

USERPTR find_nick(char *nick)
{
   User *c;
   c = userhead;

   assert(nick);
   while (c)
   {
      if (!strcasecmp(nick,c->nick)) return c;
      c = c->next;
   }
   return NULL;
}

SERVPTR find_server(char *name)
{
   Server *c;
   c = servhead;

   assert(name);
   while (c)
   {
      if (!strcasecmp(name,c->name)) return c;
      c = c->next;
   }
   return NULL;
}

Link *find_user_link(CHANPTR chan, USERPTR user)
{
   Link *lp;

   if (!user || !chan) return NULL;
   lp = chan->userhead;

   while(lp)
   {
     if (lp->value.user == user) return lp;
     lp = lp->next;
   }
   return NULL;
}

Link *find_chan_link(USERPTR user, CHANPTR chan)
{
   Link *lp;

   if (!user || !chan) return NULL;
   lp = user->chanhead;

   while(lp)
   {
     if (lp->value.chan == chan) return lp;
     lp = lp->next;
   }
   return NULL;
}

int is_chanop(USERPTR user, char *name)
{
   CHANPTR chan;
   Link *ptr;

   if (!(chan = find_channel(name))) return 0;
   else if (!(ptr = find_chan_link(user,chan))) return 0;
   else if (IsOped(ptr)) return 1;
   else return 0;
}

void purge_user(USERPTR user)
{
   Link *lp;

//   debug("Split","User %s split with server %s",user->nick,user->server->name);

   for (lp = user->chanhead; lp; lp = lp->next)
   {
      delete_chan_user(lp->value.chan,user);
      if (!lp->value.chan->userhead) delete_channel(curchan);
   }
   while (user->chanhead) delete_user_chan(user,user->chanhead->value.chan);

   delete_servuser(user->server,user);
   delete_user(user);
}

void cleanup_users(SERVPTR serv)
{
   while (serv->userhead) purge_user(serv->userhead->value.user);
}
