#ifndef STRUCT_H_
#define STRUCT_H_

#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include <stdlib.h>
#include <unistd.h>
#include <assert.h>
#include <time.h>
#include <signal.h>
#include <sys/time.h>
#include "defines.h"
#include "config.h"

typedef struct _connection Connection;
typedef struct _channel Channel;
typedef struct _user User;
typedef struct _server Server;
typedef struct _link Link;
typedef struct _services Services;
typedef struct _nickuser Nickuser;
typedef struct _chanuser Chanuser;
typedef Connection *CONNPTR;
typedef Channel *CHANPTR;
typedef User *USERPTR;
typedef Server *SERVPTR;
typedef Services *SERVICES;
typedef Nickuser *NICKUSER;
typedef Chanuser *CHANUSER;

#include "proto.h"

#ifdef HASHWATCH
struct myhash
{
  unsigned int connections;
  unsigned int channels;
  unsigned int users;
  unsigned int servers;
  unsigned int links;
  unsigned int services;
  unsigned int nickuser;
  unsigned int chanuser;
} myhash;
#endif /* HASHWATCH */

struct _connection
{
     struct _connection *next, *prev;
     int                fd;
     char               buffer[512];
     unsigned short     bufindex;
     unsigned short     type;
     unsigned short     status;
     time_t             lasttry;
};

struct _channel
{
     time_t             timestamp;              /* Channel timestamp */
     char               *name;                  /* Channel name (malloc'ed) */
     unsigned int       modes;                  /* Channel modes */

     Link               *bans;                  /* Bans */
     Link               *exinvite;              /* Invite Exemptions */
     Link               *exban;                 /* Ban Exemptions */
     Link               *userhead;              /* Channel Users */

     struct _channel    *next, *prev;
};

struct _user
{
     time_t             timestamp;              /* TS on Nick */
     time_t             last_nickreg;           /* Last nick register */
     time_t             last_chanreg;           /* Last chan register */
     time_t             signon;                 /* When the user logged on */
     char               nick[NICKLEN + 1];      /* Users Nickname */
     char               user[USERLEN + 1];      /* Users Ident (malloc later?) */
     char               host[HOSTLEN + 1];      /* Users Host (malloc later?) */
     char               *info;                  /* Users info */
     unsigned short     attempts;               /* Password Attempts */
     unsigned int       flags;                  /* User Modes */

     Link               *chanhead;              /* User Channel list */
     Server             *server;                /* Server User is on */

     struct _user       *next, *prev;
};

struct _server
{
     char               name[HOSTLEN + 1];     /* Server name (malloc later?) */
     char               *info;                 /* Server info */
     int                hops;                  /* Hop count */

     Link               *down;                 /* Server Down links */
     Link               *userhead;             /* Server's userlist */

     struct _server     *next, *prev, *up;
};

struct _services
{
     char               *nick;                 /* Service Bot nick */
     char               *user;                 /* Service Bot user */
     char               *host;                 /* Service Bot host */
     char               *info;                 /* Service Bot info */
     char               *modes;                /* Service Bot modes */
     unsigned short     type;                  /* Type of Service Bot */
 
     struct _services   *next, *prev;
};

struct _nickuser
{
     char    		*nick;                 /* Nickserv user's nick (malloc'ed) */
     char               pass[PASSLEN + 1];     /* User's password (encrypted) */
     char 		*lastmask;             /* User's last hostmask */
     char		*lastinfo;             /* User's last info */
     char               **access;              /* User's access list */
     time_t		created;               /* TS when user was registered */
     time_t		lastonline;            /* TS when user was last online */
     unsigned int       modes;                 /* User modes */
     unsigned short     accesscount;           /* Number of host's in access list */
     unsigned short     channelcount;          /* Number of channels has registered */

     Link               *channels;             /* Channels user has access to */
     struct _nickuser	*next, *prev;
};

struct _chanuser
{
     char               *name;                 /* Channel's name (malloc'ed) */
     char               *desc;                 /* Channel's desc (malloc'ed) */
     char               *url;                  /* Channel's URL (malloc'ed)  */
     char               pass[PASSLEN + 1];     /* Channel's password (encrypted) */
    
     time_t             registered;            /* When channel was registered */
     time_t             lastused;              /* Last time channel was used */

     Link               *accesslist;           /* Channel's access list */
     Nickuser           *founder;              /* Pointer to the founder */
     Nickuser           *successor;            /* Pointer to the sucessor (NULL if none) */
     Nickuser           *lastidentify;         /* Pointer to the last user to identify */

     struct _chanuser   *next, *prev;
};

struct _link
{
  struct _link *next, *prev;
  union
  {
     Server *serv;
     Channel *chan;
     User *user;
     struct
     {
       char *hostmask;
       char *who;
     } ban;
  } value;
  unsigned int modes;
};

#endif
